package com.jdbernard.timeanalyzer.categories

import com.jdbernard.timeanalyzer.events.Event
import org.joda.time.Duration
import org.joda.time.format.PeriodFormat
import org.joda.time.format.PeriodFormatter

/**
 * A `Category` represents a collection of like `Events` and sub-categories.
 */
public abstract class Category implements Comparable<Category> {

    public static PeriodFormatter periodFormatter = PeriodFormat.getDefault()

    /** List of events directly under this category. */
    public List events

    /** List of sub-categories under this category. */
    public List categories

    /** List of `CategorizationPlan`s to use when adding new `Event`s to the
     * category. */
    public List categorizationPlans

    /** A end-user-friendly text description of the category.*/
    public String description

    public Category() { this("Unamed Category") }

    public Category(String description) {
        events = []
        categories = []
        categorizationPlans = []
        this.description = description }

    /**
     * Does the given event belong in this category?
     * @param e `Event` being considered.
     * @return **`true`** if this event belongs in this category, **`false`**
     *         otherwise.
     */
    public abstract boolean matchesEvent(Event e)

    /**
     * Add an event to this category. This method does not check to see if the
     * `Event` matches this category. It assumed that the check has already been
     * made and the `Event` matches.*/
    public Event addEvent(Event event) {

        def addedEvent

        // Try first to add it to a subcategory (or create a new subcategory).
        addedEvent = addToSubcategory(event)
        
        // Cannot add to a subcategory, add to ourselves.
        if (!addedEvent) {
            events << event
            addedEvent = event }

        return addedEvent }

    public Event addToSubcategory(Event e) {

        // find all matching subcategories
        def matchingCategories = categories.findAll { it.matchesEvent(e) }

        if (matchingCategories) {
            return matchingCategories[0].addEvent(e) }

        // no matching subcategories, can we create a new one based on one
        // of our plans?
        def matchingPlans = categorizationPlans.findAll {
            it.deservesNewCategory(e, events) }

        if (matchingPlans) {
            // create the new category
            def newCategory = matchingPlans[0].newCategory(e, events)

            // add it to our list of cateogries
            categories << newCategory

            // add the new event to the category
            def addedEvent = newCategory.addEvent(e)

            // move all the events that match the new category over
            def existingEvents = matchingPlans[0].findEventsToRecategorize(e, events)
            events -= existingEvents
            existingEvents.each { newCategory.addEvent(it) }
            
            // return the new entry
            return addedEvent }

        return null }

    public Category filter(List<CategoryFilter> filters) {

        // create new filtered category
        FilteredCategory fc = new FilteredCategory(description)
        fc.filters = filters

        // filter all events and add them to the category
        fc.events

        // TODO
    }

    public Category filter(CategoryFilter filter) { return filter([filter]) }

    public Category filter(Closure c) { return filter(c as CategoryFilter) }

    public Duration getDuration() { 
        return categories.sum(new Duration(0)) { it.duration } +
               events.sum(new Duration(0)) { it.duration } }

    public int compareTo(Category other) {
        return this.getDuration().compareTo(other.getDuration()) }

    public String toString() {
        String period = periodFormatter.print(this.duration.toPeriod())
        return "${description} (${period})" }
}
